using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Threading;
using System.IO;

namespace RakeTendril
{
    /// <summary>
    /// talks to Max/MSP Jitter
    /// </summary>
    class Communicator
    {
        public volatile bool NewImageWaiting;
        public long FramesSent;
        public volatile bool Stopped;

        private Thread thread;
        private string ipAddress;
        private int port;
        private MainForm mainForm;

        public Communicator(string ipAddress, int port, MainForm mainForm)
        {
            this.ipAddress = ipAddress;
            this.port = port;
            this.mainForm = mainForm;
            thread = new Thread(threadFunction);
            thread.Start();
        }

        public void Stop()
        {
            thread.Abort();
        }

        private void threadFunction()
        {
            const int bytesPerPixel = 3;
            int byteCount = MainForm.bitmapWidth * MainForm.bitmapHeight * bytesPerPixel;
            int xStride = bytesPerPixel;
            int yStride = MainForm.bitmapWidth * bytesPerPixel;
            int[] header = new int[]{74,77,84,88,32,1,0,0,88,84,77,74,0,0,1,32,0,0,0,bytesPerPixel,
                                    0,0,0,0,0,0,0,2,0,0,MainForm.bitmapWidth/256,MainForm.bitmapWidth%256,0,0,MainForm.bitmapHeight/256,MainForm.bitmapHeight%256,
                                    0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,
                                    0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,
                                    0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,
                                    0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,
                                    0,0,0,1,0,0,0,1,
                                    xStride>>24,(xStride&0xFF0000)>>16,(xStride&0xFF00)>>8,xStride&0xFF,
                                    yStride>>24,(yStride&0xFF0000)>>16,(yStride&0xFF00)>>8,yStride&0xFF,
                                    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                                    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                                    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                                    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                                    0,0,0,0,0,0,0,0,
                                    byteCount>>24,(byteCount&0xFF0000)>>16,(byteCount&0xFF00)>>8,byteCount&0xFF,
                                    65,51,220,176,158,39,90,248 // these 8 byte of time data will be filled in dynamically
                                };
            byte[] headerBytes = new byte[header.Length];
            for (int i = 0; i < header.Length; i++)
            {
                headerBytes[i] = (byte)header[i];
            }

            TcpClient client = null;
            Stream stream = null;
            MemoryStream auxStream = new MemoryStream(8);
            BinaryWriter bw = new BinaryWriter(auxStream);
            try
            {
                client = new TcpClient(ipAddress, port);
                stream = client.GetStream();
                byte[] imageBuffer = new byte[byteCount];
                int bufSize = client.ReceiveBufferSize / 4;
                byte[] dummyBuffer = new byte[bufSize];

                DateTime start = DateTime.Now;

                while(client.Connected)
                {
                    if (!NewImageWaiting)
                    {
                        Thread.Sleep(2);
                        continue;
                    }

                    lock (mainForm.bitmap)
                    {
                        NewImageWaiting = false;
                        unsafe
                        {
                            BitmapData bp = mainForm.bitmap.LockBits(new Rectangle(0, 0, MainForm.bitmapWidth, MainForm.bitmapHeight),
                                ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb);
                            int i = 0;
                            byte* data = (byte*)bp.Scan0;
                            for (int y = 0; y < MainForm.bitmapHeight; y++)
                            {
                                byte* data1 = data;
                                for (int x = 0; x < 3*MainForm.bitmapWidth; x++) //3* : RGB
                                {
                                    imageBuffer[i] = *data1;
                                    i++;
                                    data1++;
                                }
                                data += bp.Stride;
                            }
                            mainForm.bitmap.UnlockBits(bp);
                        }
                    }

                    double elapsed = 1000.0*(DateTime.Now - start).TotalSeconds;
                    auxStream.Position = 0;
                    bw.Write(elapsed);
                    auxStream.Position = 0;
                    for (int i = 0; i < 8; i++)
                    {
                        headerBytes[headerBytes.Length-i-1] = (byte)auxStream.ReadByte();
                    }
                    stream.Write(headerBytes, 0, header.Length);
                    stream.Write(imageBuffer, 0, byteCount);

                    // ignore what we get back
                    while (client.Available >= bufSize)
                    {
                        stream.Read(dummyBuffer, 0, bufSize);
                    }

                    System.Threading.Interlocked.Increment(ref FramesSent);
                }
            }
            catch (Exception)
            {}
            finally
            {
                if (auxStream != null)
                {
                    auxStream.Close();
                }
                if (bw != null)
                {
                    bw.Close();
                }
                if (stream != null)
                {
                    stream.Close();
                }
                if (client != null)
                {
                    client.Close();
                }
                Stopped = true;
            }
        }
    }
}
